#include <unistd.h>
#include <getopt.h>
#include <dirent.h>
#include <stdexcept>
#include <string.h>
#include <errno.h>
#include <fstream>
#include <iostream>
#include <mutex>

#include "SourceFile.h"
#include "BBBParser.h"
#include "parallel.h"
#include "toolchain.h"
#include "globals.h"

static void readDir(std::string dir, std::vector<SourceFile> &sources,
		BBBParser::Variables vars);

static void readPath(std::string dir, std::string name,
		std::vector<SourceFile> &sources, BBBParser::Variables vars) {
	std::string path = dir + "/" + name;
	sys::FileInfo finfo = sys::fileInfo(path);

	if (finfo.isDir) {
		// We don't want to send 'files'
		BBBParser::Variables subvars = vars;
		subvars.erase("files");
		readDir(path, sources, std::move(subvars));
	} else {
		SourceFile::FileType type = SourceFile::fileTypeFrom(name);
		if (type != SourceFile::FileType::UNKNOWN) {
			sources.emplace_back(dir, name, type, vars);
		}
	}
}

static void readDir(std::string dir, std::vector<SourceFile> &sources,
		BBBParser::Variables vars) {
	std::ifstream buildFile(dir + "/build.bbb");
	std::vector<std::string> files;
	bool hasFiles = false;

	// Parse $dir/build.bbb, specifies a 'files' value
	if (buildFile.good()) {
		BBBParser parser(buildFile, BBBParser::FLAG_NONE);
		parser.parse(vars);

		auto it = vars.find("files");
		if (it != vars.end()) {
			files = it->second;
			hasFiles = true;
		}
	}

	// If build.bbb didn't specify 'files', we have to readdir
	if (!hasFiles) {
		DIR *d = opendir(dir.c_str());
		if (d == NULL) {
			throw std::runtime_error("opendir '" + dir + "': " + strerror(errno));
			return;
		}

		struct dirent *ent;
		while ((ent = readdir(d)) != NULL) {
			if (ent->d_name[0] == '.')
				continue;
			files.emplace_back(ent->d_name);
		}

		closedir(d);
	}

	// Go through files
	for (auto &ent: files) {
		std::string path = dir + "/" + ent;
		BBBParser::Variables subvars = vars;
		readPath(dir, ent, sources, vars);
	}
}

static void printState(const std::vector<SourceFile> &sources) {
	for (auto &source: sources) {
		std::cout << source.dir() << '/' << source.name() << ":\n";
		for (auto &kv: source.vars()) {
			std::cout << "    " << kv.first << ":\n";
			for (auto &val: kv.second) {
				std::cout << "        " << val << '\n';
			}
		}
	}
}

static bool compile(const std::vector<SourceFile> &sources, const std::string &outDir, int jobs) {
	bool compiled = false;
	std::mutex mut;
	parallel::parallel(jobs, sources, [&](const SourceFile &source) {
		if (source.needsRecompile(outDir)) {
			source.compile(outDir);
			mut.lock();
			compiled = true;
			mut.unlock();
		}
	});

	return compiled;
}

static void link(const std::vector<SourceFile> &sources, const std::string &outDir) {
	// TODO: Actually link
	printf("Linking\n");
}

int main(int argc, char **argv) {
	std::vector<std::string> srcDirs;
	std::string outDir = "bbbuild";
	int jobs = parallel::coreCount();
	std::string workDir = "";

	enum class Action {
		BUILD, PRINT_STATE,
	};
	Action action = Action::BUILD;

	const char *shortopts = "hvo:j:C:p";
	const struct option opts[] = {
		{ "help",        no_argument,       NULL, 'h' },
		{ "verbose",     no_argument,       NULL, 'v' },
		{ "output",      required_argument, NULL, 'o' },
		{ "jobs",        required_argument, NULL, 'j' },
		{ "directory",   required_argument, NULL, 'C' },
		{ "print-state", no_argument,       NULL, 'p' },
		{},
	};

	const char usage[] =
		"Usage: bbbuild [options...] [sources]\n"
		"\n"
		"  -h, --help             "
			"Show this help text.\n"
		"  -v, --verbose          "
			"Show every command as it's executing."
		"  -o, --output <dir>     "
			"Set output directory. Default: bbbuild\n"
		"  -j, --jobs <count>     "
			"Set the number of jobs run simultaneously. "
			"Default: the number of cores in the machine.\n"
		"  -C, --directory <dir>  "
			"Change directory before doing anything else.\n"
		"  -p, --print-state      "
			"Print the state instead of building.\n";

	// Parse options from argv
	while (1) {
		int optidx;
		int c = getopt_long(argc, argv, shortopts, opts, &optidx);

		if (c < 0) {
			break;
		}

		switch (c) {
			case 'h':
				puts(usage);
				return 0;

			case 'v':
				global::verbose = true;
				break;

			case 'o':
				outDir = optarg;
				break;

			case 'j':
				jobs = atoi(optarg);
				if (jobs <= 0) {
					fprintf(stderr, "Can't run %i jobs.\n", jobs);
					return 1;
				}
				break;

			case 'C':
				workDir = optarg;
				break;

			case 'p':
				action = Action::PRINT_STATE;
				break;

			default:
				printf("Unknown option: '%c'.\n%s", (char)c, usage);
				return 1;
		}
	}

	// Change directory?
	if (workDir.size() > 0) {
		if (global::verbose) {
			fprintf(stderr, "Entering directory '%s'\n", workDir.c_str());
		}

		if (chdir(workDir.c_str()) < 0) {
			perror(workDir.c_str());
			return 1;
		}
	}

	// Parse non-opt argv as source dirs
	if (optind < argc) {
		while (optind < argc) {
			srcDirs.push_back(argv[optind++]);
		}
	} else {
		srcDirs.push_back(".");
	}

	// Read configs from source dirs
	std::vector<SourceFile> sources;
	for (std::string &dir: srcDirs) {
		readDir(dir, sources, {});
	}

	switch (action) {
	case Action::BUILD:
		if (compile(sources, outDir, jobs)) {
			link(sources, outDir);
		} else {
			std::cerr << "Nothing to do.\n";
		}

		break;

	case Action::PRINT_STATE:
		printState(sources);
		break;
	}
}
